#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "copyrigh.h"
#include "wattcp.h"
#include "misc.h"
#include "udp_dom.h"
#include "pcbuf.h"
#include "pctcp.h"
#include "wdpmi.h"     /* for DOS4GW/WDOSX */
#include "bsdname.h"

int _getpeername (const tcp_Socket *s, void *dest, int *len)
{
  struct watt_sockaddr temp;
  int    ltemp;

  memset (&temp, 0, sizeof(temp));
  temp.s_ip   = s->hisaddr;
  temp.s_port = s->hisport;

  if (!s->hisaddr || !s->hisport || !_chk_socket((sock_type*)s))
  {
    if (len)
       *len = 0;
    return (-1);
  }

  /* how much do we move
   */
  ltemp = (len ? *len : sizeof (struct watt_sockaddr));
  if (ltemp > sizeof (struct watt_sockaddr))
      ltemp = sizeof (struct watt_sockaddr);
  memcpy (dest, &temp, ltemp);

  if (len)
     *len = ltemp;
  return (0);
}

int _getsockname (const tcp_Socket *s, void *dest, int *len)
{
  struct watt_sockaddr temp;
  int    ltemp;

  memset (&temp, 0, sizeof(temp));
  temp.s_ip   = s->myaddr;
  temp.s_port = s->myport;

  if (!s->hisaddr || !s->hisport || !_chk_socket((sock_type*)s))
  {
    if (len)
       *len = 0;
    return (-1);
  }

  /* how much do we move
   */
  ltemp = (len ? *len : sizeof (struct watt_sockaddr));
  if (ltemp > sizeof (struct watt_sockaddr))
      ltemp = sizeof (struct watt_sockaddr);
  memcpy (dest, &temp, ltemp);

  if (len)
     *len = ltemp;
  return (0);
}

/*
 * Return the host's domain name.
 * Set errno on failure and return -1.
 */
int getdomainname (char *name, int length)
{
  int slen = strlen (def_domain);

  if (!def_domain || !*def_domain || length <= slen)
  {
    errno = EINVAL;
    return (-1);
  }
#if (DOSX)
  if (!valid_addr((DWORD)name, slen+1))
  {
    errno = EFAULT;
    return (-1);
  }
#endif
  strcpy (name, def_domain);
  return (0);
}

/*
 * Set the host's domain name.
 * Set errno on failure and return -1.
 */
int setdomainname (const char *name, int len)
{
  if (len < 0 || len > sizeof(defaultdomain)-1)
  {
    errno = EINVAL;
    return (-1);
  }
#if (DOSX)
  if (!valid_addr((DWORD)name, len))
  {
    errno = EFAULT;
    return (-1);
  }
#endif
  def_domain = memcpy (defaultdomain, name, len);
  def_domain[len] = 0;
  return (0);
}

/*
 * Make a FQDN from `hostname' & `def_domain'.
 * Set errno on failure and return -1.
 */                                                   
int gethostname (char *name, int len)
{
  int req_len;         /* required length for supplied buffer */
  int has_domain = 0;  /* our hostname contains a domain part */

#if (DOSX)
  if (!valid_addr((DWORD)name, len))
  {
    errno = EFAULT;
    return (-1);
  }
#endif

  if (!hostname[0])
  {
    *name = 0;
    return (0);   /* !! is this an error? */
  }

  req_len = strlen (hostname);
  if (def_domain && *def_domain)
  {
    req_len += 1 + strlen (def_domain);
    has_domain = 1;
  }

  if (len < req_len)
  {
    errno = EINVAL;
    return (-1);
  }

  strcpy (name, hostname);
  if (has_domain)
  {
    strcat (name, ".");
    strcat (name, def_domain);
  }
  return (0);
}

/*
 * Expects a "Fully Qualified Domain Name" in `fqdn'.
 * Split at first `.' and extract `hostname' and `def_domain'.
 * Set errno on failure and return -1.
 */
int sethostname (const char *fqdn, int len)
{
  const char *dot;

#if (DOSX)
  if (!valid_addr((DWORD)fqdn, len))
  {
    errno = EFAULT;
    return (-1);
  }
#endif

  if (!fqdn || !*fqdn || len < 0 || len > MAX_HOSTLEN)
  {
    errno = EINVAL;
    return (-1);
  }
  dot = strchr (fqdn, '.');
  if (!dot)
  {
    strcpy (hostname, fqdn);
    return (0);
  }
  strncpy (hostname, fqdn, (size_t)(dot-fqdn));
  len -= (int)(dot-fqdn);
  return setdomainname (dot+1, len);
}


/*
 * From Ralf Brown's interrupt list
 *
 * --------D-215E00-----------------------------
 * INT 21 - DOS 3.1+ network - GET MACHINE NAME
 *         AX = 5E00h
 *         DS:DX -> 16-byte buffer for ASCII machine name
 * Return: CF clear if successful
 *             CH = validity
 *                 00h name invalid
 *                 nonzero valid
 *                     CL = NetBIOS number for machine name
 *                     DS:DX buffer filled with blank-paded name
 *         CF set on error
 *             AX = error code (01h) (see #1366 at AH=59h)
 * Note:   supported by OS/2 v1.3+ compatibility box, PC-NFS
 */

/* Copyright (C) 1997 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */

/*
 * Try asking a LAN extension of DOS for a host-name.
 */
int _get_machine_name (char *buf, int size)
{
  char *h, dosbuf[16];
  int   len;

#if (DOSX & DJGPP)
  __dpmi_regs r;

  r.x.ax = 0x5E00;
  r.h.ch = 0;         /* Try to detect overloading of 0x5e00 */
  r.x.dx = __tb & 15;
  r.x.ds = __tb / 16;
  __dpmi_int (0x21, &r);
  if ((r.x.flags & 1) || r.h.ch == 0)
     return (-1);

  dosmemget (__tb, sizeof(dosbuf), dosbuf);

#elif (DOSX & PHARLAP)
  SWI_REGS r;

  if (_watt_dosTbSize < sizeof(dosbuf) || !_watt_dosTbr)
     return (-1);

  r.edx = RP_OFF (_watt_dosTbr);
  r.ds  = RP_SEG (_watt_dosTbr);
  r.eax = 0x5E00;
  r.ecx = 0;
  _dx_real_int (0x21, &r);
  if ((r.flags & 1) || hiBYTE(r.ecx) == 0) /* Carry set or CH == 0 */
     return (-1);

  ReadFarMem ((void*)&dosbuf, _watt_dosTbp, sizeof(dosbuf));

#elif (DOSX & (DOS4GW|WDOSX))
  struct DPMI_regs r;

  if (_watt_dosTbSize < sizeof(dosbuf) || !_watt_dosTbSeg)
     return (-1);

  r.r_ds = _watt_dosTbSeg;
  r.r_dx = 0;
  r.r_ax = 0x5E00;
  r.r_cx = 0;
  if (!dpmi_real_interrupt(0x21, &r) || hiBYTE(r.r_cx) == 0)
     return (-1);

  memcpy (&dosbuf, SEG_OFS_TO_LIN(_watt_dosTbSeg,0), sizeof(dosbuf));

#elif (DOSX & POWERPAK)
  UNFINISHED();
  strcpy (dosbuf, "random-pc");

#elif (DOSX == 0)
  union  REGS  r;
  struct SREGS s;

  r.x.ax = 0x5E00;
  r.h.ch = 0;
  r.x.dx = FP_OFF (dosbuf);
  s.ds   = FP_SEG (dosbuf);
  int86x (0x21, &r, &r, &s);
  if (r.x.cflag || r.h.ch == 0)
     return (-1);

#else
  #error Unknown target!
#endif

  /* Remove right space padding
   */
  h = dosbuf + strlen (dosbuf);
  while (h > dosbuf && h[-1] == ' ')
        h--;

  *h  = 0;
  h   = dosbuf;
  len = strlen (h);
  if (len + 1 > size)
  {
    errno = ERANGE;
    return (-1);
  }
  strcpy (buf, h);
  return (0);
}
