/*
 * IP output routines
 *
 *  Version
 *
 *  1.0 : Apr 18, 1999 : G. Vanem - created
 *  1.1 : Jan 29, 2000 : G. Vanem - added functions for
 *                       handling list of IP-fragments.
 */

#include <stdio.h>
#include <string.h>

#include "copyrigh.h"
#include "wattcp.h"
#include "chksum.h"
#include "pcsed.h"
#include "pctcp.h"
#include "pcstat.h"
#include "ip_out.h"

int _default_ttl     = 254;
int _default_tos     = 0;
int _ip_id_increment = 1;   /* some stacks use 5 */

static WORD ip_id = 0;

/*
 *  Increment IP-identifier before returning it.
 *
 *  Note: Use _get_this_ip_id() for sending IP fragments.
 */
WORD _get_ip_id (void)
{
  ip_id += _ip_id_increment;
  return intel16 (ip_id);
}

/*
 *  Return IP-identifier but don't increment it.
 */
WORD _get_this_ip_id (void)
{
  return intel16 (ip_id);
}

int _ip_output (in_Header  *ip,        /* ip-structure to fill in */
                DWORD       src_ip,    /* from address (network order!) */
                DWORD       dst_ip,    /* dest address (network order!) */
                WORD        protocol,  /* IP-protocol number */
                BYTE        ttl,       /* Time To Live */
                BYTE        tos,       /* Type Of Service, 0 unspecified */
                int         data_len,  /* length of data after ip header */
                const void *sock,      /* which socket is this */
                const char *file,      /* Debug: from what file */
                unsigned    line)      /*  and line was _ip_output called */
{
  int len = sizeof(*ip) + data_len;
 
 /* This is already done in _eth_formatpacket()
  * memset (ip,0,sizeof(*ip));
  */

  if (src_ip == 0)
      src_ip = intel (my_ip_addr);

  ip->ver            = 4;
  ip->hdrlen         = sizeof(*ip) / 4;
  ip->length         = intel16 (len);
  ip->tos            = tos;
  ip->ttl            = ttl ? ttl : _default_ttl;
  ip->proto          = protocol;
  ip->identification = _get_ip_id();
  ip->source         = src_ip;
  ip->destination    = dst_ip;
  ip->checksum       = 0;
  ip->checksum       = ~checksum (ip, sizeof(*ip));

#if defined(USE_DEBUG)
  if (_dbugxmit)
    (*_dbugxmit) (sock, ip, file, line);
#else
  ARGSUSED (sock);
  ARGSUSED (file);
  ARGSUSED (line);
#endif

  return _eth_send (len);
}

/*
 * Should belong in ip_in.c (not yet)
 *
 *  Check for correct IP-version, header length and header checksum.
 *  If failed, discard the packet. ref. RFC1122: section 3.1.2.2
 */
int _chk_ip_header (const in_Header *ip)
{
  unsigned ip_hlen = in_GetHdrLen (ip);

  if (ip->ver != 4)    /* We only speak IPv4 */
  {
    DEBUG_RX (NULL, ip);
    IP_STAT (ips_badvers++);
    return (0);
  }          
  if (ip_hlen < sizeof(*ip))
  {
    DEBUG_RX (NULL, ip);
    IP_STAT (ips_tooshort++);
    return (0);
  }          
  if (checksum(ip,ip_hlen) != 0xFFFF)
  {
    DEBUG_RX (NULL, ip);
    IP_STAT (ips_badsum++);
    return (0);
  }
  return (1);
}


#if defined(USE_FRAGMENTS)
/*
 *  Split a large buffer into multiple IP-fragments ready for
 *  transmission by _ip_output() or _eth_send().
 */
struct frag_list * ip_do_fragment (sock_type *sk, const void *buf, int len)
{
  ARGSUSED (sk);                 /* !!to-do */
  ARGSUSED (buf);
  ARGSUSED (len);
  return (NULL);
}

void ip_free_fragment (struct frag_list *frag)
{
  ARGSUSED (frag);
}
#endif

