/*
 *  This file contains code for displaying the Intel Cpu identification
 *  that has been performed by CheckCpuType() function.
 *
 *  COPYRIGHT (c) 1998 valette@crf.canon.fr
 *
 *  The license and distribution terms for this file may be
 *  found in the file LICENSE in this distribution or at
 *  http://www.OARcorp.com/rtems/license.html.
 *
 *  $Id: displayCpu.c,v 1.4 1998/08/24 16:59:20 joel Exp $
 *
 *  Heavily modified for Watt-32. Added port to Watcom-386,
 *  calculate CPU speed.  G.Vanem  <giva@bgnett.no> 2000
 */

/*
 * Tell us the machine setup..
 */
#include <stdio.h>
#include <string.h>

#include "wattcp.h"
#include "misc.h"
#include "cpu.h"

#if !defined(HAVE_UINT64)
#error I need 64-bit integer support
#endif

char Cx86_step = 0;

char *Cx86_type[] = {
     "unknown", "1.3", "1.4", "1.5", "1.6", "2.4",
     "2.5", "2.6", "2.7 or 3.7", "4.2"
   };

static char *i486model (unsigned int nr)
{
  static char *model[] = {
              "0", "DX", "SX", "DX/2", "4", "SX/2", "6",
              "DX/2-WB", "DX/4", "DX/4-WB", "10", "11", "12", "13",
              "Am5x86-WT", "Am5x86-WB"
            };
  if (nr < DIM(model))
     return (model[nr]);
  return (NULL);
}

static char *i586model (unsigned int nr)
{
  static char *model[] = {
              "0", "Pentium 60/66", "Pentium 75+", "OverDrive PODP5V83",
              "Pentium MMX", NULL, NULL, "Mobile Pentium 75+",
              "Mobile Pentium MMX"
            };
  if (nr < DIM(model))
     return (model[nr]);
  return (NULL);
}

static char *Cx86model (void)
{
  unsigned char nr6x86 = 0;
  static char *model[] = {
              "unknown", "6x86", "6x86L", "6x86MX", "MII"
            };

  switch (x86_type)
  {
    case 5:
         /* cx8 flag only on 6x86L */
         nr6x86 = ((x86_capability & (1 << 8)) ? 2 : 1);
	 break;
    case 6:
	 nr6x86 = 3;
	 break;
    default:
	 nr6x86 = 0;
  }

  /* We must get the stepping number by reading DIR1
   */
  outp (0x22, 0xff);
  x86_mask = inp (0x23);

  switch (x86_mask)
  {
    case 0x03:
	 Cx86_step = 1;		/* 6x86MX Rev 1.3 */
	 break;
    case 0x04:
	 Cx86_step = 2;		/* 6x86MX Rev 1.4 */
	 break;
    case 0x05:
	 Cx86_step = 3;		/* 6x86MX Rev 1.5 */
	 break;
    case 0x06:
	 Cx86_step = 4;		/* 6x86MX Rev 1.6 */
	 break;
    case 0x14:
	 Cx86_step = 5;		/* 6x86 Rev 2.4 */
	 break;
    case 0x15:
	 Cx86_step = 6;		/* 6x86 Rev 2.5 */
	 break;
    case 0x16:
	 Cx86_step = 7;		/* 6x86 Rev 2.6 */
	 break;
    case 0x17:
	 Cx86_step = 8;		/* 6x86 Rev 2.7 or 3.7 */
	 break;
    case 0x22:
	 Cx86_step = 9;		/* 6x86L Rev 4.2 */
	 break;
    default:
	 Cx86_step = 0;
  }
  return (model[nr6x86]);
}

static char *i686model (unsigned int nr)
{
  static char *model[] = {
              "PPro A-step", "Pentium Pro"
            };
  if (nr < DIM(model))
     return (model[nr]);
  return (NULL);
}

struct cpu_model_info {
       int   x86;
       char *model_names[16];
     };

static struct cpu_model_info amd_models[] = {
  { 4,
    { NULL, NULL, NULL, "DX/2", NULL, NULL, NULL, "DX/2-WB", "DX/4",
      "DX/4-WB", NULL, NULL, NULL, NULL, "Am5x86-WT", "Am5x86-WB"
    }
  },
  { 5,
    { "K5/SSA5 (PR-75, PR-90, PR-100)", "K5 (PR-120, PR-133)",
      "K5 (PR-166)", "K5 (PR-200)", NULL, NULL,
      "K6 (166 - 266)", "K6 (166 - 300)", "K6-2 (200 - 450)",
      "K6-3D-Plus (200 - 450)", NULL, NULL, NULL, NULL, NULL, NULL
    }
  },
};

static char *AMDmodel (void)
{
  char *p = NULL;
  int   i;

  if (x86_model < 16)
  {
    for (i = 0; i < DIM(amd_models); i++)
      if (amd_models[i].x86 == x86_type)
      {
        p = amd_models[i].model_names[(int)x86_model];
	break;
      }
  }
  return (p);
}

static char *getmodel (int x86, int model)
{
  char  *p = NULL;
  static char nbuf[12];

  if (!strncmp (x86_vendor_id, "Cyrix", 5))
     p = Cx86model();
  else if (!strcmp (x86_vendor_id, "AuthenticAMD"))
     p = AMDmodel();
  else
  {
    switch (x86)
    {
      case 4:
	   p = i486model (model);
	   break;
      case 5:
	   p = i586model (model);
	   break;
      case 6:
	   p = i686model (model);
	   break;
    }
  }
  if (p)
    return (p);

  sprintf (nbuf, "%d", model);
  return (nbuf);
}

void PrintCpuInfo (void)
{
  static char *x86_cap_flags[] = {
              "fpu", "vme", "de", "pse", "tsc", "msr", "pae", "mce",
              "cx8", "apic", "10", "11", "mtrr", "pge", "mca", "cmov",
              "16", "17", "18", "19", "20", "21", "22", "mmx",
              "24", "25", "26", "27", "28", "29", "30", "31"
            };
  int i;

  printf ("cpu      : %c86\n", x86_type + '0');
  printf ("model    : %s\n", have_cpuid ? getmodel(x86_type,x86_model) :
                                          "unknown");

  if (x86_vendor_id[0] == '\0')
     strcpy (x86_vendor_id, "unknown");

  printf ("vendor_id: %s\n", x86_vendor_id);

  if (x86_mask)
  {
    if (!strncmp (x86_vendor_id, "Cyrix", 5))
         printf ("stepping : %s\n", Cx86_type[(int)Cx86_step]);
    else printf ("stepping : %d\n", x86_mask);
  }
  else
    printf ("stepping : unknown\n");

  printf ("fpu      : %s\n", (hard_math  ? "yes" : "no"));
  printf ("cpuid    : %s\n", (have_cpuid ? "yes" : "no"));
  printf ("flags    :");

  for (i = 0; i < 32; i++)
  {
    if (x86_capability & (1 << i))
       printf (" %s", x86_cap_flags[i]);
  }
  printf ("\n");
}

/*
 * Function for obtaining current RDTSC value.
 */
#if defined(__DJGPP__)

  /*
   * Generate code for the initial timestamp read - the cld
   * and nops are included for repeatable pairing and to eliminate
   * shadowing effects from previous instructions.
   *
   * This code is originally by
   *   Tom Burgess <Tom_Burgess@bc.sympatico.ca> and
   *   Douglas Eleveld <deleveld@dds.nl>
   */
  static __inline__ uint64 get_rdtsc (void)
  {
    uint64 tsc;
    __asm__ __volatile__ (
            ".byte 0x0F, 0x31;"
            "movl %%eax, (%%edi);"
            "movl %%edx, 4(%%edi);"
            "cld;"
            "nop; nop; nop;"
            "nop; nop; nop;"
            "nop; nop;"
            :: "D" (&tsc) : "eax", "edx");
    return (tsc);
  }
  static __inline__ void get_rdtsc2 (uint64 *tsc)
  {
    __asm__ __volatile__ (
            ".byte 0x0f, 0x31"  /* rdtsc opcode */
            : "=a" (tsc[0]), "=d" (tsc[1]) : );
}


#elif defined(__WATCOM386__)
  extern uint64 get_rdtsc (void);
  #pragma aux get_rdtsc = \
          ".586"          \
          "db 0Fh, 31h"   \
          "cld"           \
          "nop"           \
          "nop"           \
          "nop"           \
          "nop"           \
          modify [eax edx];
#else
  #error Unsupported compiler
#endif

/*
 * Return estimated CPU speed in MHz
 */
double get_cpu_speed (uint64 *clocks_per_msec)
{
  #define NUM_MSEC     100  /* 100msec should be enough */
  #define NUM_SAMPLES  3

  double mhz = 0.0;
  double sample [NUM_SAMPLES];
  int    i;

  if ((x86_capability & 0x10) == 0 ||  /* rdtsc not supported */
      !has_8254)                       /* no 8254 PIT !? */
     return (mhz);

  for (i = 0; i < NUM_SAMPLES; i++)
  {
    DWORD  t = set_timeout (NUM_MSEC);
    uint64 start, stop;

    start = get_rdtsc();
    while (!chk_timeout(t))
          ;
    stop = get_rdtsc();
    sample[i] = (stop - start) / (1000.0*NUM_MSEC);
  }

  mhz = 0.0;
  for (i = 0; i < NUM_SAMPLES; i++)
      mhz += (double)sample[i];
  mhz /= (uint64)NUM_SAMPLES;

  if (clocks_per_msec)
     *clocks_per_msec = (uint64) (mhz/1000.0);
  return (mhz);
}

#ifdef TEST_PROG

int main (void)
{
  double mhz;

#ifdef USE_EXCEPT
  #include <exc.h>
  InstallExcHandler (NULL);
#endif

  init_misc();
  CheckCpuType();
  PrintCpuInfo();

  mhz = get_cpu_speed (NULL);
  if (mhz)
       printf ("clock    : %.3f MHz\n", mhz);
  else printf ("clock    : RDTSC not supported\n");
  return (0);
}
#endif

