#include <stdio.h>
#include <string.h>
#include <ctype.h>

#define BYTE  unsigned char
#define DWORD unsigned long

#if defined(__GNUC__)
void get_cpuid (DWORD idval, DWORD *eax, DWORD *ebx, DWORD *ecx, DWORD *edx)
{
  __asm__ __volatile__ (
         ".byte 0x0f,0xa2"   /* cpuid opcode */
         : "=a" (*eax), "=b" (*ebx), "=c" (*ecx), "=d" (*edx)
         : "0" (idval));
}

#elif defined(__WATCOMC__) && defined(__386__)  /* wcc386 */
extern void get_cpuid (DWORD, DWORD *eax, DWORD *ebx, DWORD *ecx, DWORD *edx);
#pragma aux get_cpuid =     \
        ".586"              \
        "push eax"          \
        "push ebx"          \
        "push ecx"          \
        "push edx"          \
        "mov eax, esi"      \
        "cpuid"             \
        "mov esi, [esp]"    \
        "mov [esi], edx"    \
        "mov esi, [esp+4]"  \
        "mov [esi], ecx"    \
        "mov esi, [esp+8]"  \
        "mov [esi], ebx"    \
        "mov esi, [esp+12]" \
        "mov [esi], eax"    \
        "add esp, 16"       \
        parm [esi] [eax] [ebx] [ecx] [edx] \
        modify [esi eax ebx ecx edx];
#endif

void print_reg (DWORD reg, char *what)
{
  BYTE a = (BYTE)(reg & 255);
  BYTE b = (BYTE)(reg >>  8)  & 255;
  BYTE c = (BYTE)(reg >> 16)  & 255;
  BYTE d = (BYTE)(reg >> 24);

  printf ("%s: %08lX, %c%c%c%c\n", what, reg,
          isprint(a) ? a : '.',
          isprint(b) ? b : '.',
          isprint(c) ? c : '.',
          isprint(d) ? d : '.');
}

int main (void)
{
  DWORD eax, ebx, ecx, edx;

  printf ("cpuid,0 (name)\n");
  get_cpuid (0, &eax, &ebx, &ecx, &edx);
  print_reg (eax, "EAX");
  print_reg (ebx, "EBX");
  print_reg (edx, "EDX");
  print_reg (ecx, "ECX");

  printf ("\ncpuid,1 (family)\n");
  get_cpuid (1, &eax, &ebx, &ecx, &edx);
  print_reg (eax, "EAX");
  print_reg (ebx, "EBX");
  print_reg (ecx, "ECX");
  print_reg (edx, "EDX");

  printf ("\ncpuid, 80000002h (AMD)\n");
  get_cpuid (0x80000002, &eax, &ebx, &ecx, &edx);
  print_reg (eax, "EAX");
  print_reg (ebx, "EBX");
  print_reg (ecx, "ECX");
  print_reg (edx, "EDX");

  printf ("\ncpuid, 80000003h (AMD)\n");
  get_cpuid (0x80000003, &eax, &ebx, &ecx, &edx);
  print_reg (eax, "EAX");
  print_reg (ebx, "EBX");
  print_reg (ecx, "ECX");
  print_reg (edx, "EDX");

  return (0);
}

