#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <unistd.h>
#include <dos.h>
#include <tcp.h>

int test_timers (DWORD msec)
{
  int i;

  printf ("system ref. (now) = %lu\n"
          "set_timeout(100)  = %lu\n",
          set_timeout(0), set_timeout(100));

  for (i = 0; i < 10; i++)
  {
    DWORD t = set_timeout (msec);

    fputc ('.', stderr);
    while (!chk_timeout(t))
          ;
  }
  return (0);
}

int test_intel (void)
{
  printf ("intel  (11223344h) = %lX\n", intel(0x11223344L));
  printf ("intel16(1144h)     = %04X\n",intel16(0x1144));
  return (0);
}

double timeval_diff (struct timeval *a, struct timeval *b)
{
  struct timeval d_tv;

  d_tv.tv_sec  = a->tv_sec  - b->tv_sec;
  d_tv.tv_usec = a->tv_usec - b->tv_usec;
  return (double)d_tv.tv_sec + 1E-6 * (double)d_tv.tv_usec;
}

int test_gettod2 (DWORD msec)
{
  struct timeval tv, last;
  DWORD  to, loop = 0;

  while (!kbhit())
  {
    gettimeofday2 (&tv, NULL);
    printf ("tod: %6d.%06ld", tv.tv_sec, tv.tv_usec);

    to = set_timeout (msec);
    while (!chk_timeout(to))
          ;

    if (loop++ > 0)
         printf (", diff: %.6fs\n", timeval_diff(&tv,&last));
    else putchar ('\n');

    last.tv_sec  = tv.tv_sec;
    last.tv_usec = tv.tv_usec;
  }
  return (0);
}

/*
 * test timers around midnight.
 * Set time to 23:59:45, set timer to timeout at 00:00:15 (30s).
 * Check for timeout. Restore time.
 */
int test_rollover (void)
{
  struct dostime_t new, old;
  DWORD  t, tnow, tprev = 0;

  _dos_gettime (&old);
  new.hour    = 23;
  new.minute  = 59;
  new.second  = 45;
  new.hsecond = 0;
  _dos_settime (&new);       /* set to 23:59:45 */

  t = set_timeout (30000);
  printf ("timeout at timer-count %lu (00:00:15)\n", t);

  while (!kbhit())
  {
    struct dostime_t now;
    int    expired;
    long   time_diff1;
    DWORD  time_diff2;

    sleep (1);              /* wait 1 sec */
    tnow = set_timeout (0);
    _dos_gettime (&now);

    expired = (chk_timeout (t) != 0);
    if (tprev)
    {
      time_diff1 = (long) (tnow - tprev);
      time_diff2 = get_timediff (tnow, tprev);
    }
    else
    {
      time_diff1 = 0;
      time_diff2 = 0;
    }

    printf ("%10lu, td1 %10ld, td2 %10lu, %02d:%02d:%02d: %c\n",
            tnow, time_diff1, time_diff2,
            now.hour, now.minute, now.second,
            expired ? '+' : '-');
    tprev = tnow;
  }
  _dos_settime (&old);      /* ignore elapsed time */
  return (0);
}

void Usage (void)
{
  puts ("Usage: ttime [-h] <-r | -g | -t | -i> \n"
        "  -h : don't use hi-resolution timers\n"
        "  -r : test timer day-rollover\n"
        "  -g : test gettimeofday2()\n"
        "  -t : test general timers\n"
        "  -i : test intel() functions\n");
  exit (-1);
}


int main (int argc, char **argv)
{
  if (argc < 2)
     Usage();

  init_misc();
  switch (getopt(argc, argv, "hrgti"))
  {
    case 'h': hires_timer (0);
    case 'r': exit (test_rollover());
    case 'g': exit (test_gettod2(1000));
    case 't': exit (test_timers(1000));
    case 'i': exit (test_intel());
    default : Usage();
  }
  return 0;
}

