/*
 * listener.c -- a datagram sockets "server" demo
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <arpa/inet.h>
#include <tcp.h>

#define MYPORT 6543    /* the port users will be connecting to */

#define MAXBUFLEN 100

int main (void)
{
     struct sockaddr_in my_addr;    /* my address information */
     struct sockaddr_in their_addr; /* connector's address information */
     int    sockfd;
     int    addr_len, numbytes;
     char   buf[MAXBUFLEN];
     
     dbug_init();
     sock_init();
     
     if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
	  perror ("socket");
	  return (1);
     }
     
     memset (&my_addr,0,sizeof(my_addr));  /* zero the rest of the struct */
     my_addr.sin_family = AF_INET;         /* host byte order */
     my_addr.sin_port = htons(MYPORT);     /* short, network byte order */
     my_addr.sin_addr.s_addr = INADDR_ANY; /* automatically fill with my IP */
     
     if (bind(sockfd, (struct sockaddr *)&my_addr, sizeof(struct sockaddr)) == -1) {
	  perror ("bind");
	  return (1);
     }

     printf ("Waiting for UDP messages at port %d...", MYPORT);
     fflush (stdout);

     for (;;) {
     
	  //  --- Wait for UDP message
	  
	  addr_len = sizeof (struct sockaddr);
	  numbytes = recvfrom (sockfd, buf, MAXBUFLEN, 0,
			       (struct sockaddr *)&their_addr, &addr_len);
	  if (numbytes < 0) {
	       perror ("recvfrom");
	       return (1);
	  }
	  
	  printf ("got packet from %s\n",inet_ntoa(their_addr.sin_addr));
	  printf ("packet is %d bytes long\n",numbytes);
	  buf[numbytes] = '\0';
	  printf ("packet contains \"%s\"\n",buf);
	  
	  //  --- Reply to this message (echo)
	  
	  numbytes = sendto (sockfd, buf, numbytes, 0,
			     (struct sockaddr*)&their_addr, sizeof(struct sockaddr));
	  
	  if (numbytes < 0) {
	       perror ("recvfrom");
	       return (1);
	  }
	  
	  printf ("sent %d bytes to %s\n", numbytes, inet_ntoa(their_addr.sin_addr));
     }
     
     close (sockfd);
     
     return (0);
}
